! SketchyLISP -- An interpreter for a purely applicative dialect of Scheme
! Copyright (C) 2005 Nils M Holm
! Derived from ArrowLISP, Copyright (C) 1998-2005 Nils M Holm.
!
! 2005-07-05
!	- Fixed quotation of arguments to continuations.
! 2005-06-29
!	- Improved some error messages.
!	- Increased MAXLOADL from 12 to 16.
! 2005-06-28
!	- Removed function tracking (:F) flag (always ON now).
!	- Removed implode numbers (:I) flag (always ON now).
! 2005-06-27
!	- Added CHAR? and STRING? primitives.
!	- Made SYMBOL? a primitive.
!	- Fixed a GC leak in TOPLOOP(). I wonder how /that/ one
!	  managed to sneak in.
! 2005-06-26
!	- Added CHAR->INTEGER and INTEGER->CHAR primitives.
!	- Added INTEGER->LIST and LIST->INTEGER primitives.
!	- CAR and CDR no longer operate on numbers. Use above
!	  primitives instead.
! 2005-06-25
!	- Added #<string> data type.
!	- Added STRING->SYMBOL and SYMBOL->STRING primitives.
!	- Added LIST->STRING and STRING->LIST primitives.
!	- Removed IMPLODE and EXPLODE primitives.
! 2005-06-24
!	- Added #<char> tag and data type.
!	- Added #<string> tag.
! 2005-06-23
!	- Removed EVAL primitive.
!	- Added #<primitive> and primitive procedure type.
!	- Moved evaluation of all special forms to EVAL().
! 2005-06-20
!	- Changed flag values for large environments, because
!	  C99 treats 0x80000000 as a /signed/ integer on 32-bit
!	  architectures, thereby breaking the C back end.
!	  So much for compatibility.
! 2005-06-15
!	- Replaced trace prefix with '+'.
!	- Replaced error message prefix with '*'.
!	- Changed error message format.
! 2005-06-14
!	- Made definitions of digit (0-DIGIT, ...) internal.
!	- Made above definitions constant, so that 0-DIGIT => '0-DIGIT
!	  rather than => '0. This way confusion of digits and numbers
!	  becomes impossible. Eg, (CDR 0) => '(0-DIGIT).
! 2005-06-02
!	- Made references to unbound symbols illegal.
!	- Removed *SYMBOLS*.
! 2005-05-30
!	- Moved tail call optimization (TCO) to separate procedure
!	  TAILCALL().
!	- Made TCO cover LETREC frames, so that the following
!	  expression (finally!) runs in constant space:
!	  (LETREC ((O (LAMBDA () (LETREC ((FOO 'BAR)) (O))))) (O))
! 2005-05-29
!	- Replaced LETREC* with LETREC.
! 2005-05-28
!	- Added atomic argument lists: (LAMBDA X X).
!	- Removed LIST primitive.
! 2005-05-26
!	- Made PROCEDURE? return #T for primitives.
! 2005-05-25
!	- Made type checking more strict. See TAGGED() and
!	  LAZYATOM() functions. E.g. (CAR (LAMBDA (X) X)) no
!	  longer works.
! 2005-05-24
!	- Added the #<NUMBER> tag.
!	- Added the NUMBER? primitive.
!	- Made PAIR? return #F for lists (sic!) beginning with
!	  #<NUMBER>.
! 2005-05-23
!	- Fixed creation of closures from lambda functions with
!	  dotted argument lists (dotted argument was not bound).
!	- Replaced (#<CONT> ...) with #<CONTINUATION>.
!	- Made (LAMBDA (X) T) replace itself with #<CLOSURE (X) ...>.
!	- Added PROCEDURE? primitive.
!	- Made PAIR? return #F for lists (sic!) beginning with
!	  #<CONTINUATION> or #<CLOSURE>.
!	- Made READ() report unreable objects starting with '#<'.
! 2005-05-22
!	- Fixed a bug in PRINT that could crash the interpreter
!	  when testing for numbers in improper lists.
! 2005-05-21
!	- Added variable arguments using improper argument lists.
!	- Implemented APPLY.
!	- Made :T applicable to primitives.
! 2005-04-29 -
! 2005-05-11
!	Created initial version
!	Differences between ArrowLISP and SketchyLISP:
!	- Extended syntax check of LAMBDA to cover third argument
!	- Dropped dynamic scoping; removed :X meta command
!	- Dropped DEFUN in favour of DEFINE w/ LAMBDA
!	- Removed NIL symbol
!	- Replaced ATOM with PAIR?
!	- Renamed *T* to #T
!	- Renamed *F* to #F
!	- Renamed EQ to EQ?
!	- Renamed NULL to NULL?
!	- Renamed LABEL to LETREC*
!	- COND no longer accepts NIL instead of #F
!	- Renamed CONTINUATION #<CONT>
!	- Bound CONTINUATION to #<CONT>
!	- F'G' now evaluates to 'F 'G (was 'F'G)
!	- DEFINE now creates dynamic closures
!	- Removed symbol T
!	- Removed recursive LAMBDA option
!	- Added second (descriptive) argument to :L and :C, eg
!	  :L LIB/ONEP.L ONE?
!	- Dropped ArrowLISPs #-Synax
!	- Made [+-]?[0-9]+ an alias of ArrowLISPs '#[+-]?[0-9]+
!	  so that numeric literals no longer require a prefix.
!	- Made (cons 1 2) => (1.2) although 2 is a list internally.

decl prznum(2), pr(1), nl(0), fatal(1);

class counter()

 var	Nred, Nr1k, Nr1m, Nr1g;

 ! Reset counter.
 public reset() do
	Nred := 0;
	Nr1k := 0;
	Nr1m := 0;
	Nr1g := 0;
 end

 ! Increment counter.
 public count(k) do
	Nred := Nred+k;
	if (Nred >= 1000) do
		Nred := Nred - 1000;
		Nr1k := Nr1k + 1;
		if (Nr1k >= 1000) do
			Nr1k := 0;
			Nr1m := Nr1m+1;
			if (Nr1m >= 1000) do
				Nr1m := 0;
				Nr1g := Nr1g+1;
				if (Nr1g >= 1000)
					fatal("statistics: counter overflow");
			end
		end
	end
 end

 ! Print counter value.
 public printvalue(s) do
	if (Nr1g) do prznum(Nr1g, 0); pr(","); end
	if (Nr1m) do prznum(Nr1m, Nr1g->3:0); pr(","); end
	if (Nr1k) do prznum(Nr1k, Nr1m->3:0); pr(","); end
	prznum(Nred, Nr1k->3:0);
	pr("\s");
	pr(s);
	nl();
 end

end

module sketchy(t3x, string, char, counter, util);

object	t[t3x], str[string], chr[char], u[util];

const	MAJORVERSION	= 3;
var	RELEASE;

const	NBASEREG =	9;	! Number of Base Registers
const	MAXWRAPCOL =	255;	! Max. wrap column

! Maximum number of nested load commands
! (Recommended: 16-bit: 12, >=32-bit 16)
const	MAXLOADL =	12;	! SMALL-ENVIRONMENT
!const	MAXLOADL =	16;	! LARGE-ENVIRONMENT

! Maximum length for symbols and imploded lists
const	TEXTLEN =	128;	! SMALL-ENVIRONMENT
!const	TEXTLEN =	1024;	! LARGE-ENVIRONMENT

! Number of nodes.
! (Recommended: 16-bit: 12280, >=32-bit: 131072)
const	POOLSIZE =	12280;	! SMALL-ENVIRONMENT
!const	POOLSIZE =	131072;	! LARGE-ENVIRONMENT

!	The NODE structure has a CAR and a CDR field.
!	The use of each field is as follows:
!	+---+---+--------------------------+
!	| M | A |   14-bit Address Field   |
!	+---+---+--------------------------+
!	The sign bit M is used to tag used nodes during
!	the mark step of the garbage collection.
!	The A field is used to tag atoms: if A=1 in a
!	CAR field, the node containing the field represents
!	a character (used to name atoms). In this case,
!	the CAR field contains the character itself in the
!	LSB and A=1 in the MSB. CDR points to the next
!	character or NIL. A is unused in CDR fields.
!	The node pool is kept in two separate vectors
!	called 'Car' and 'Cdr'. Consequently, Car[i] is
!	used to access the CAR field of the i'th node in
!	the pool and Cdr[i] is used to access its CDR
!	field.
!	The pointer reversing collector uses CDRs M flag
!	to store the state of an already visited atom.

! Masks for the 'A' and 'M' flags as desribed above.
const	AFLAG = 0x4000,		! SMALL-ENVIRONMENT
	MFLAG = 0x8000;		! SMALL-ENVIRONMENT
!const	AFLAG = 0x20000000,	! LARGE-ENVIRONMENT
!	MFLAG = 0x40000000;	! LARGE-ENVIRONMENT

const	EOF	=	%1;	! EOF indicator
const	DOT	=	%2;	! Internal: dot character
const	RPAREN	=	%3;	! Internal: right parenthesis

var	NIL;			! Not In List (or Pool)
var	Car[POOLSIZE],		! Vector holding CAR fields
	Cdr[POOLSIZE];		! Vector holding CDR fields
var	Infile;			! Input file name
var	Buffer,			! Ptr to current buffer
	Bufvec::MAXLOADL*TEXTLEN+MAXLOADL,	! Load buffers
	Bp, Ep,			! Input pointer, limit
	Input;			! Current input stream
var	Rejected;		! Unread character
var	Mstack, Lstack;		! Mode stack, List stack
var	Bstack;			! Binding stack, used by LETREC
var	Estack;			! Environment stack, for fixing closures
var	Parent;			! Parent pointer used in D/S/W GC
var	Free;			! Freelist
var	Symbols;		! Symbol table
var	Safesymbols;		! Safe copy of symbols
var	Stack, Stack0;		! Global stack, bottom of Stack
var	Frame;			! Current call frame
var	Tmp, Tmp2;		! Safe locations
var	Base[NBASEREG];		! Base registers (roots for GC)
var	Level;			! Nesting level during input
var	Loadlev;		! Nesting level of LOAD meta cmds
var	Evlev;			! Number of nested EVALs
var	Errflag;		! Error flag
var	Function;		! Name of current lambda function
var	Trace;			! Name of traced function
var	Prtbuf::MAXWRAPCOL+1,	! Print buffer,
	Pp;			! Pointer
var	Quoted;			! Quote flag of PRINT
var	Nlseq::3, Nll;		! Newline sequence, length
var	Wrapcol;		! Wrap column
var	Prefix;			! Output line prefix
var	Maxuse, Lastuse;	! Memory use gauge
var	Line;			! Input line number
object	Reductions[counter],	! Reduction counter
	Allocations[counter];	! Allocation counter
var	Statflag;		! Statistics flag
var	Ntrace;			! Max functions to print in call trace
var	Prcllevl;		! Print closure body flag
var	Lexenv;			! Environment for creating closures
var	Bound;			! Variables bound in a closure
var	Image::TEXTLEN;		! Name of initial load image

! Builtin symbol pointers (for fast lookup)
var	S_apply, S_callcc, S_char, S_closure, S_cond, S_continuation,
	S_define, S_false, S_lambda, S_letrec, S_number, S_primitive,
	S_quote, S_string, S_t, S_true, S_void, S_last, S_0, S_1, S_2,
	S_3, S_4, S_5, S_6, S_7, S_8, S_9;

! Command line flags
var	O_batch;	! Batch (non-interactive) mode

! Primitive function opcodes
struct N_PRIMITIVES =
	P_BOTTOM, P_CAR, P_CDR, P_CHAR_TO_INTEGER, P_CHARP, P_CONS,
	P_EQP, P_INTEGER_TO_CHAR, P_INTEGER_TO_LIST, P_LIST_TO_INTEGER,
	P_LIST_TO_STRING, P_NULLP, P_NUMBERP, P_PAIRP, P_PROCEDUREP,
	P_STRING_TO_LIST, P_STRING_TO_SYMBOL, P_STRINGP,
	P_SYMBOL_TO_STRING, P_SYMBOLP;

! Primitive function pointers.
var	Primitives[N_PRIMITIVES];

! Emit a newline sequence
newline() t.write(T3X.SYSOUT, Nlseq, Nll);

! Print the string S thru a buffered interface.
pr(s) do var k, n, lim;
	if (\s) do
		if (Prefix /\ Wrapcol)
			t.write(T3X.SYSOUT, Prefix, str.length(Prefix));
		t.write(T3X.SYSOUT, PrtBuf, Pp);
		Pp := 0;
		return 0;
	end
	k := str.length(s);
	lim := Wrapcol-> Wrapcol: MAXWRAPCOL;
	while (Pp + k > lim) do
		if (Prefix /\ Wrapcol)
			t.write(T3X.SYSOUT, Prefix, str.length(Prefix));
		n := lim - Pp;
		t.memcopy(@Prtbuf::Pp, s, n);
		t.write(T3X.SYSOUT, Prtbuf, lim);
		Pp := 0;
		k := k-n;
		s := @s::n;
		if (Wrapcol) newline();
	end
	if (k) do
		t.memcopy(@Prtbuf::Pp, s, k);
		Pp := Pp+k;
	end
end

! Print a padded number with leading characters.
_prnum(n, w, spaces) do var b::20, k;
	k := str.length(str.numtostr(b, n, 10));
	if (k < w) pr(@spaces::(k));
	pr(b);
end

! Print number with leading spaces.
prnum(n, w) _prnum(n, w, "\s\s\s\s\s\s");

! Print number with leading zeroes.
prznum(n, w) _prnum(n, w, "000");

! Flush the output buffer and emit a newline sequence.
nl() do
	pr(0);
	newline();
end

! Convert string to lower case.
locase(s) do var k, i;
	k := str.length(s)+1;
	for (i=0, k)
		if ('A' <= s::i /\ s::i <= 'Z')
			s::i := s::i + ('a' - 'A');
	return s;
end

decl print(1), _print(1);

! Print function names on call stack
tracecall() do var s, v, n;
	s := Frame;
	n := Ntrace;
	while (s \= NIL) do
		if (\n \/ Cdr[s] = NIL \/ Car[Cdr[s]] = NIL) leave;
		if (n = Ntrace) pr("* Trace:");
		n := n-1;
		pr(" ");
		Quoted := 1;
		_print(Car[Cdr[s]]);
		s := Car[s];
	end
	if (n \= Ntrace) nl();
end

! Print error message M and set Errflag.
! If N \= %1 print an additional line containing '* N'.
error(m, n) do
	if (Errflag) return NIL;
	Prefix := "* ";
	pr("* ");
	if (Infile) do
		pr(locase(Infile));
		pr(": ");
	end
	ie (Function \= NIL) do
		Quoted := 1;
		_print(Function);
	end
	else do
		pr("Top-level");
	end
	pr("[");
	prnum(Line, 0);
	pr("]: ");
	pr(m);
	if (n \= %1) do
		if (m::0) pr(": ");
		Quoted := 1;
		_print(n);
	end
	nl();
	if (Frame \= NIL) tracecall();
	if (O_batch) halt 1;
	Errflag := %1;
	Function := NIL;
	Prefix := 0;
	return NIL;
end

! Print message M and halt the interpreter.
fatal(m) do
	error(m, %1);
	pr("* Fatal error, aborting");
	nl();
	halt 1;
end

! Mark nodes which can be accessed through N.
! This routine uses the Deutsch/Schorr/Waite algorithm
! (aka pointer reversal algorithm) which marks the
! nodes of a pool in constant space.
! In addition to the Mark flag (MFLAG), it uses
! another bit to save the state of the current
! node (S). The (otherwise unused) MFLAG of CDR is
! used to store the state.
! Each visited node goes through these states:
! M=0 S=0 unvisited, process CAR
! M=1 S=1 CAR visited, process CDR
! M=1 S=0 completely visited, return to parent
mark(n) do var p;
	Parent := NIL;	! Initially, there is no parent node
	while (1) do
		! Reached a leaf?
		ie (n = NIL \/ Car[n] & MFLAG) do
			! When the current node is a leaf and there is
			! no parent, the entire tree is marked.
			if (Parent = NIL) leave;
			ie (Cdr[Parent] & MFLAG) do
				! State 2: the CDR of the parent has
				! not yet been marked (M of CDR set).
				! Swap CAR and CDR pointers and
				! proceed with CDR. Set State=3.
				p := Cdr[Parent] & ~MFLAG;
				Cdr[Parent] := Car[Parent] & ~MFLAG;	! S=0
				Car[Parent] := n | MFLAG;		! M=1
				n := p;
			end
			else do
				! State 3: CAR and CDR of parent done.
				! Return to the parent and restore
				! parent of parent.
				p := Parent;
				Parent := Cdr[p];
				Cdr[p] := n;
				n := p;
			end
		end
		else do
			! State 1: The current node has not yet been
			! visited.
			ie (Car[n] & AFLAG) do
				! If the node is an atom, go directly
				! to state 3: Save the parent in CDR,
				! make the current node the new parent
				! and move to its CDR.
				p := Cdr[n];
				Cdr[n] := Parent;			! S=0
				Parent := n;
				n := p;
				Car[Parent] := Car[Parent] | MFLAG;	! M=1
			end
			else do
				! Go to state 2: like above, but save
				! the parent in CAR and proceed to CAR.
				p := Car[n];
				Car[n] := Parent | MFLAG;		! M=1
				Parent := n;
				n := p;
				Cdr[Parent] := Cdr[Parent] | MFLAG;	! S=1
			end
		end
	end
end

! Mark and Sweep Garbage Collection.
! First, mark all nodes which can be accessed through
! Base Registers (Base[]) and then reclaim untagged
! nodes.
gc() do var i, n, k;
	k := 0;
! do pr("GC called"); nl(); end
	for (i=0, NBASEREG) mark(Base[i][0]);
	Free := NIL;
	for (i=0, POOLSIZE) do
		ie (\(Car[i] & MFLAG)) do
			Cdr[i] := Free;
			Free := i;
			k := k+1;
		end
		else do
			Car[i] := Car[i] & ~MFLAG;
		end
	end
	if (Maxuse < POOLSIZE-k) Maxuse := POOLSIZE-k;
! prnum(k, 0); pr(" nodes reclaimed"); nl();
	return k;
end

! Allocate a fresh node and initialize with PCAR,PCDR.
alloc(pcar, pcdr) do var n;
	if (Statflag) Allocations.count(1);
	if (Free = NIL) do
		gc();
		if (Free = NIL) fatal("out of nodes");
	end
	n := Free;
	Free := Cdr[Free];
	Car[n] := pcar;
	Cdr[n] := pcdr;
	return n;
end

! Save node N on the Stack.
save(n) do
	Tmp := n;	! Otherwise, alloc() might recycle this node.
	Stack := alloc(n, Stack);
	Tmp := NIL;
end

! Pop K nodes off the Stack and return
! the most recently popped one.
unsave(k) do var n;
	while (k) do
		if (Stack = NIL) fatal("stack underflow");
		n := Car[Stack];
		Stack := Cdr[Stack];
		k := k-1;
	end
	return n;
end

! Save value V on the M-Stack.
msave(v) do
	! Since the Mstack holds integer values rather than
	! nodes, the values are packaged in the character
	! fields of atoms.
	Car[Mstack] := alloc(v|AFLAG, Car[Mstack]);
end

! Pop a value off the M-Stack and return it.
munsave() do var v;
	if (Car[Mstack] = NIL) fatal("m-stack underflow");
	v := Car[Car[Mstack]] & ~AFLAG;		! See msave()
	Car[Mstack] := Cdr[Car[Mstack]];
	return v;
end

! Save node N on the L-Stack.
lsave(n) do
	Tmp := n;	! Otherwise, alloc() might recycle this node.
	Lstack := alloc(n, Lstack);
	Tmp := NIL;
end

! Pop K nodes off the L-Stack and return
! the most recently popped one.
lunsave(k) do var n;
	while (k) do
		if (Lstack = NIL) fatal("l-stack underflow");
		n := Car[Lstack];
		Lstack := Cdr[Lstack];
		k := k-1;
	end
	return n;
end

! Save node N on the B-Stack.
bsave(n) do
	Tmp := n;	! Otherwise, alloc() might recycle this node.
	Bstack := alloc(n, Bstack);
	Tmp := NIL;
end

! Pop K nodes off the B-Stack and return
! the most recently popped one.
bunsave(k) do var n;
	while (k) do
		if (Bstack = NIL) fatal("b-stack underflow");
		n := Car[Bstack];
		Bstack := Cdr[Bstack];
		k := k-1;
	end
	return n;
end

! Read a single character from the input stream
! and return it. Rdch()=EOF indicates that the
! input is exhausted.
! Characters will be stripped to seven bits.
_rdch() do var c;
	if (Rejected \= EOF) do
		c := Rejected;
		Rejected := EOF;
		return c;
	end
	if (Bp >= Ep) do
		Ep := t.read(Input, Buffer, TEXTLEN);
		Bp := 0;
	end
	if (Ep < 1) return EOF;
	Bp := Bp+1;
	c := Buffer::(Bp-1) & 127;
	if (c = '\n') Line := Line+1;
	return c;
end

! Read a character and convert it to lower case.
rdch() do var c;
	c := _rdch();
	return ('A' <= c /\ c <= 'Z')-> c-'A'+'a' : c;
end

! Find a symbol named S in the symbol table.
! Each symbol is represented by a (NAME.VALUE) pair
! where NAME is a list of character nodes and value
! may be any valid S-expression.
! The symbol table is a list containing symbol
! pairs ((N1.V1) ...).
! When a symbol named S is found, return its
! pair (S.V) and otherwise return NIL.
findsym(s) do var y, n, i;
	y := Symbols;
	while (y \= NIL) do
		n := Car[Car[y]];
		i := 0;
		while (n \= NIL /\ s::i) do
			if (s::i \= Car[n] & 255) leave;
			n := Cdr[n];
			i := i+1;
		end
		if (n = NIL /\ \s::i) return Car[y];
		y := Cdr[y];
	end
	return NIL;
end

! Implode a string.
implodestr(s) do var i, n, m, a;
	i := 0;
	if (s::i = 0) return NIL;
	n := NIL;
	while (s::i) do
		m := alloc(s::i | AFLAG, NIL);
		ie (n = NIL) do		! Protect the first character
			n := m;
			save(n);
		end
		else do		! Just append the rest
			Cdr[a] := m;
		end
		a := m;
		i := i+1;
	end
	unsave(1);
	return n;
end

! Add the symbol S to the symbol table if it
! does not already exist. If it does exist,
! return the existing symbol.
! When adding a new symbol, initialize the
! VALUE field with V. If V=0, make the symbol
! a constant (bind it to itself).
! Return the pair representing the symbol S.
addsym(s, v) do var n, m;
	n := findsym(s);
	if (n \= NIL) return n;
	n := implodestr(s);
	save(n);
	m := alloc(n, v-> v: n);
	save(m);
	Symbols := alloc(m, Symbols);
	unsave(2);
	return m;
end

! Add primitive procedure.
addprim(name, opcode) do var y;
	y := addsym(name, 0);
	Cdr[y] := alloc(S_primitive, NIL);
	Cdr[Cdr[y]] := alloc(opcode | AFLAG, NIL);
	Cdr[Cdr[Cdr[y]]] := y;
	return y;
end

decl	read(0);

! Read a list (S0 ... SN) and return (a pointer to) it.
! This routine also recognizes pairs of the form (S0.S1).
! For empty lists, it returns NIL.
rdlist() do
	var	n,	! Node read
		m,	! Ptr to list
		a,	! Used to append nodes to m
		c;	! Member counter
	var	badpair;

	badpair := "bad pair";
	Level := Level+1;
	m := alloc(NIL, NIL);	! Root node
	save(m);
	a := NIL;
	c := 0;
	while (1) do
		n := read();
		if (n = EOF) fatal("EOF in list");
		if (n = DOT) do
			if (c < 1) do
				error(badpair, %1);
				loop;
			end
			n := read();
			Cdr[a] := n;
			if (read() \= RPAREN) do
				error(badpair, %1);
				loop;
			end
			unsave(1);
			Level := Level-1;
			return m;
		end
		if (n = RPAREN) leave;
		ie (a = NIL) 
			a := m;		! First member: insert at root
		else
			a := Cdr[a];	! Following members: append
		Car[a] := n;
		Cdr[a] := alloc(NIL, NIL);	! Alloc space for next member
		c := c+1;
	end
	Level := Level-1;
	if (a \= NIL) Cdr[a] := NIL;	! Remove trailing empty node
	unsave(1);
	return c-> m: NIL;
end

! Read characters from the input device until a non-space
! character [^\t\s] is found. Return that character.
skip() do var c;
	c := rdch();
	while (c = '\s' \/ c = '\t') c := rdch();
	return c;
end

imagevars() return [
	@Statflag, @Prcllevl, @Wrapcol,
	@Symbols, @Free,
	@S_apply, @S_callcc, @S_char, @S_closure, @S_cond, @S_continuation,
	@S_define, @S_false, @S_letrec, @S_lambda, @S_number, @S_primitive,
	@S_quote, @S_string, @S_t, @S_true, @S_void, @S_last, @S_0, @S_1,
	@S_2, @S_3, @S_4, @S_5, @S_6, @S_7, @S_8, @S_9,
0 ];

! Dump node pool image to file named in P.
dump(p) do var fd, n, v, i, magic;
	fd := t.open(p, T3X.OWRITE);
	if (fd < 0) do
		error("cannot create file", %1);
		pr("* "); pr(p); nl();
		return NIL;
	end
	magic := "SKETCHY__";
	magic::7 := t.bpw();
	magic::8 := MAJORVERSION;
	t.write(fd, magic, 9);
	n := POOLSIZE;
	t.write(fd, @n, t.bpw());
	v := imagevars();
	i := 0;
	while (v[i]) do
		t.write(fd, v[i], t.bpw());
		i := i+1;
	end
	if (	t.write(fd, Car, POOLSIZE*t.bpw()) \= POOLSIZE*t.bpw() \/
		t.write(fd, Cdr, POOLSIZE*t.bpw()) \= POOLSIZE*t.bpw()
	) do
		t.close(fd);
		pr("dump failed"); nl();
		return;
	end
	t.close(fd);
	if (\O_batch) do
		pr("dumped"); nl();
	end
end

decl toploop(0);

! Load S-expressions from the external file or device
! named in the string P. Return nothing.
load(p) do
	var	o, of, buf, b, e, r;
	var	fd, oline;

	if (Loadlev+1 >= MAXLOADL) return error("load level too deep", %1);
	fd := t.open(p, T3X.OREAD);
	if (fd < 0) do
		error("cannot open file", %1);
		pr("* "); pr(p); nl();
		return NIL;
	end
	Loadlev := Loadlev + 1;
	! Save old I/O state
	buf := Buffer;
	b := Bp;
	e := Ep;
	r := Rejected;
	! Reset I/O buffer
	Bp := 0;
	Ep := 0;
	Rejected := EOF;
	Buffer := @Bufvec::((TEXTLEN+1)*Loadlev);
	! Run the toplevel loop with redirected I/O
	o := Input;
	Input := fd;
	oline := Line;
	Line := 1;
	of := Infile;
	Infile := p;
	toploop();
	Infile := of;
	Line := oline;
	! Restore previous I/O state
	Rejected := r;
	Ep := e;
	Bp := b;
	Buffer := buf;
	Loadlev := Loadlev - 1;
	Input := o;
	t.close(fd);
	if (Level) error("unbalanced parentheses in loaded file", %1);
	if (\Loadlev /\ \O_batch) do
		pr("loaded"); nl();
	end
end

! Read all characters up to the next line terminator (or the EOF)
! from the input device and store them in the vector P.
! Convert each character read to lower case.
! Return the terminating character.
! This routine is used to accept parameters to meta commands.
getname(p) do var c, i;
	c := skip();
	i := 0;
	while (c \= '\s' /\ c \= '\t' /\ c \= '\n' /\ c \= '\r' /\ c \= EOF)
	do
		! Lowcase characters to avoid
		! SUCH_NAMES under Unix.
		if ('A' <= c /\ c <= 'Z') c := c-'A'+'a';
		p::i := c;
		if (i >= TEXTLEN-2) do
			error("name too long", %1);
			return 0;
		end
		c := rdch();
		i := i+1;
	end
	p::i := 0;
	return c;
end

! Extract a YES/NO, ON/OFF, 0/1 argument used to
! enable and disable options.
yesno(p, pf) do var c, f;
	c := getname(p);
	f := 2;
	if (p::0 = 'y' \/ p::0 = '1') f := %1;
	if (p::0 = 'n' \/ p::0 = '0') f := 0;
	if (p::0 = 'o') do
		if (p::1 = 'n') f := %1;
		if (p::1 = 'f') f := 0;
	end
	if (f = 2) error("Usage: command on/off", %1);
	pf[0] := f;
	return c;
end

! Count the nodes in the freelist.
countfree() do var n, k;
	k := 0;
	n := Free;
	while (n \= NIL) do
		k := k+1;
		n := Cdr[n];
	end
	return k;
end

! Print a line followed by a newline sequence.
pl(s) do
	pr(s); nl();
end


! Print the conditions of use.
license() do var w;
	w := Wrapcol;
	Wrapcol := 79;
pl(
"SketchyLISP -- An Interpreter for Purely Symbolic LISP");
pl(
"Copyright (C) 2005 Nils M Holm.  All rights reserved.");
pl("");
pl(
"Redistribution and use in source and binary forms, with or without");
pl(
"modification, are permitted provided that the following conditions");
pl(
"are met:");
pl(
"1. Redistributions of source code must retain the above copyright");
pl(
"   notice, this list of conditions and the following disclaimer.");
pl(
"2. Redistributions in binary form must reproduce the above copyright");
pl(
"   notice, this list of conditions and the following disclaimer in the");
pl(
"   documentation and/or other materials provided with the distribution.");
pl("");
pl(
"THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND");
pl(
"ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE");
pl(
"IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE");
pl(
"ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE");
pl(
"FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL");
pl(
"DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS");
pl(
"OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)");
pl(
"HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT");
pl(
"LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY");
pl(
"OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF");
pl(
"SUCH DAMAGE.");
pl("");
	Wrapcol := w;
end

! Normal termination
bye() do
	if (\O_batch) do
		pr("Bye"); nl();
	end
	halt;
end

! Set and acknowledge state of flag
setflag(name, varp, val) do
	if (val \= 2) do
		varp[0] := val;
		if (\Loadlev /\ \O_batch) do
			pr(name);
			pl(varp[0]-> " on": " off");
		end
	end
end

! Read and process meta (colon) commands.
meta() do
	var	c, k, k2, y;
	var	p::TEXTLEN+1;

	c := skip();
	ie (c = 'c') do			! Load conditionally
		c := getname(p);
		k := str.scan(p, '/')+1;
		k2 := str.scan(p, '.');
		if (k2 \= %1) p::k2 := 0;
		y := findsym(locase(@p::k));
		if (k2 \= %1) p::k2 := '.';
		ie (y = NIL \/ Cdr[y] = S_void) do
			load(locase(p));
		end
		else if (\Loadlev /\ \O_batch) do
			pr("found"); nl();
		end
	end
	else ie (c = 'd') do		! Dump node pool image
		c := getname(p);
		dump(p);
	end
	else ie (c = 'g') do		! Trigger GC
		k := countfree();
		k2 := gc();
		if (Maxuse \= 0) Lastuse := Maxuse;
		Maxuse := 0;
		prnum(Lastuse, 6);
		pr(" nodes used max."); nl();
		prnum(k2-k, 6);
		pr(" nodes reclaimed"); nl();
		prnum(k2, 6);
		pr(" nodes in freelist"); nl();
	end
	else ie (c = 'k') do		! Switch closure printing on/off
		c := getname(p);
		k := str.strtonum(p, 10, 0);
		if (k > 2) k := 2;
		Prcllevl := k>2-> 2: k;
		pr("closure print level = ");
		prnum(Prcllevl, 0); nl();
	end
	else ie (c = 'l') do		! Load program
		c := getname(p);
		load(p);
	end
	else ie (c = 'q') do		! Quit
		bye();
	end
	else ie (c = 's') do		! Switch statistics mode on/off
		c := yesno(p, @k);
		setflag("statistics", @Statflag, k);
	end
	else ie (c = 't') do		! Specify/reset function to trace
		c := getname(p);
		Trace := findsym(p);
		ie (Trace \= NIL) do
			if (\O_batch) do
				pr("tracing "); pr(p); nl();
			end
		end
		else ie (p::0) do
			error("no such symbol", %1);
			pr("* "); pr(p); nl();
		end
		else do
			if (\O_batch) do
				pr("trace off"); nl();
			end
		end
	end
	else ie (c = 'v') do		! Version info or License
		c := getname(p);
		ie (p::0) do
			license();
		end
		else do
			pr("SketchyLISP (C) 2005 Nils M Holm"); nl();
			pr("Latest modification: ");
			pr(RELEASE); nl();
			prnum(POOLSIZE, 6);
			pr(" total nodes"); nl();
		end
	end
	else ie (c = 'w') do		! Specify wrap column
		c := getname(p);
		k := str.strtonum(p, 10, 0);
		ie (\k) do
			Wrapcol := 0;
			if (\O_batch) do
				pr("wrap=off"); nl();
			end
		end
		else ie (k < 1 \/ k > MAXWRAPCOL) do
			error("bad wrap column", %1);
		end
		else do
			Wrapcol := k;
			if (\O_batch) do
				pr("wrap="); prnum(Wrapcol, 0); nl();
			end
		end
	end
	else ie (c = '#') do		! Reset line number
		Line := 0;
	end
	else do
		pr("Usage: :c file  - load file conditionally"); nl();
		pr("       :d file  - dump image to file"); nl();
		pr("       :g       - collect garbage"); nl();
		pr("       :k 0/1/2 - set closure print level"); nl();
		pr("       :l file  - load file"); nl();
		pr("       :q       - quit"); nl();
		pr("       :s 0/1   - statistics on/off"); nl();
		pr("       :t name  - trace function (:t=off)"); nl();
		pr("       :v[l]    - show version/license"); nl();
		pr("       :w col   - set wrap column (0=none)"); nl();
	end
	while (c \= '\n' /\ c \= EOF) c := rdch();
end

! Is N a 'real' (non-NIL) Atom?
atomic(n) return n \= NIL /\ Car[n] \= NIL /\ Car[Car[n]] & AFLAG;

! Is N a tagged list (an internal type)?
tagged(n) return n \= NIL /\ Car[n] \= S_true /\ Car[n] \= S_false /\
		Car[Car[Car[n]]] & ~AFLAG = '#';

! Is N a lazy atom (atomic or tagged list)?
lazyatom(n) return atomic(n) \/ tagged(n);

! Quote an expression
quote(n) do var q;
	save(n);
	q := alloc(S_quote, NIL);
	save(q);
	Cdr[q] := alloc(n, NIL);
	unsave(2);
	return q;
end

! Check whether a string represents a number.
! Numbers are defined as [+-]?[0-9]+.
numericstr(s) do var i;
	i := 0;
	if (s::0 = '+' \/ s::0 = '-')
		i := 1;
	if (\s::i) return 0;
	while (s::i) do
		if (\chr.digit(s::i)) return 0;
		i := i+1;
	end
	return 1;
end

! Explode a string into a list of one-character symbols.
explodenum(s) do var i, l, x, name, y;
	i := 0;
	l := alloc(S_number, NIL);
	x := l;
	save(l);
	name := "0-digit";
	while (s::i) do
		name::0 := s::i;
		name::1 := chr.digit(s::i)-> '-': 0;
		y := addsym(name, NIL);
		Cdr[x] := alloc(y, NIL);
		x := Cdr[x];
		i := i+1;
	end
	unsave(1);
	return l;
end

! Report unreadable object
unreadable() do var c, i, buf::2;
	error("unreadable object", %1);
	pr("* #<");
	buf::1 := 0;
	while (1) do
		c := rdch();
		if (c = '>' \/ c = '\n') leave;
		buf::0 := c;
		pr(buf);
	end
	pr(">"); nl();
	return NIL;
end

! Create a character literal.
mkchar(x) do var ch;
	Tmp2 := alloc(S_char, NIL);
	Cdr[Tmp2] := alloc(x | AFLAG, NIL);
	ch := Tmp2;
	Tmp2 := NIL;
	return ch;
end

! Read a character literal.
character() do var buf::10, i, c;
	for (i=0, 9) do
		c := _rdch();
		if (i > 0 /\ \chr.alpha(c)) leave;
		buf::i := c;
	end
	Rejected := c;
	buf::i := 0;
	ie (i = 0) c := '\s';
	else ie (i = 1) c := buf::0;
	else ie (\str.comp(buf, "space")) c := '\s';
	else return error("bad character name", %1);
	return mkchar(c);
end

! Read a string literal.
stringliteral() do var s::TEXTLEN, c, i, n, q;
	i := 0;
	q := 0;
	c := _rdch();
	while (q \/ c \= '"') do
		if (i >= TEXTLEN-2) do
			error("symbol too long", %1);
			i := i-1;
		end
		s::i := c;
		q := c = '\\';
		if (\q) i := i+1;
		c := _rdch();
	end
	s::i := 0;
	n := implodestr(s);
	save(n);
	n := alloc(n, NIL);
	n := alloc(S_string, n);
	unsave(1);
	return n;
end

! Read a symbol or a numeric literal. When reading a
! symbol, add it to the global symbol table.
sym_or_num(c) do var s::TEXTLEN, i;
	i := 0;
	while (c \= '\s' /\ c \= '\t' /\ c \= '\n' /\
		c \= '\r' /\ c \= '(' /\ c \= ')' /\
		c \= ';' /\ c \= '.' /\ c \= '#' /\
		c \= '\''
	) do
		if (i >= TEXTLEN-2) do
			error("symbol too long", %1);
			i := i-1;
		end
		s::i := c;
		i := i+1;
		c := rdch();
	end
	s::i := 0;
	Rejected := c;
	if (numericstr(s)) return explodenum(s);
	return addsym(s, S_void);
end

! Read an expression from the current input stream
! and return (a pointer to) it.
! Each valid expression is either an Atom or a List.
read() do var c;
	c := rdch();
	while (1) do	! Skip spaces and comments, run meta commands
		while (c = '\s' \/ c = '\t' \/ c = '\n' \/ c = '\r')
			c := rdch();
		if (c = ':' /\ \Level) do
			meta();
			c := '\n';
			Errflag := 0;
			loop;
		end
		if (c \= ';') leave;
		while (c \= '\n') c := rdch();
	end
	if (c = EOF) return EOF;
	ie (c = '(') do
		return rdlist();
	end
	else ie (c = '\'') do
		return quote(read());
	end
	else ie (c = '#') do
		c := rdch();
		if (c = 'f') return S_false;
		if (c = 't') return S_true;
		if (c = '\\') return character();
		if (c = '<') return unreadable();
		return error("bad # syntax", %1);
	end
	else ie (c = '"') do
		return stringliteral();
	end
	else ie (c = ')') do
		if (\Level) return error("unexpected ')'", %1);
		return RPAREN;
	end
	else ie (c = '.') do
		if (\Level) return error("unexpected '.'", %1);
		return DOT;
	end
	else do
		return sym_or_num(c);
	end
end

! Misc. error reporting and handling routines...

wrongargs(n) return error("wrong argument count", n);

badargl(n) return error("bad argument list", n);

! Evaluate N=(CONS M M2)
docons(n) do var m, m2;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] = NIL \/ Cdr[Cdr[m]] \= NIL)
		return wrongargs(n);
	m2 := Car[Cdr[m]];
	m := alloc(Car[m], m2);
	return m;
end

! Evaluate N=(CAR M)
docar(n) do var m;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	if (atomic(Car[m]) \/ Car[m] = NIL \/ tagged(Car[m]))
		return error("non-pair in 'car'", Car[m]);
	return Car[Car[m]];
end

! Evaluate N=(CDR M)
docdr(n) do var m;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	if (atomic(Car[m]) \/ Car[m] = NIL \/ tagged(Car[m]))
		return error("non-pair in 'cdr'", Car[m]);
	m := Cdr[Car[m]];
	return m;
end

! Evaluate N=(CHAR->INTEGER M)
dochartointeger(n) do var m, b::4, i, c;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	m := Car[m];
	if (atomic(m) \/ m = NIL \/ Car[m] \= S_char)
		return error("non-character in 'char->integer'", m);
	c := Car[Cdr[m]] & ~AFLAG;
	i := 4;
	b::i := 0;
	while (c \/ i = 4) do
		i := i-1;
		b::i := c mod 10 + '0';
		c := c / 10;
	end
	return explodenum(@b::i);
end

! Evaluate N=(CHAR? M)
docharp(n) do var m;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	m := Car[m];
	return \atomic(m) /\ Car[m] = S_char -> S_true: S_false;
end

! Evaluate N=(INTEGER->CHAR M)
dointegertochar(n) do var m, c, p;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	m := Car[m];
	if (atomic(m) \/ m = NIL \/ Car[m] \= S_number)
		return error("non-integer in 'integer->char'", m);
	p := Cdr[m];
	c := 0;
	while (p \= NIL) do
		c := c * 10 + (Car[Car[Car[p]]] & ~AFLAG) - '0';
		p := Cdr[p];
	end
	if (c .> 127)
		return error("value out of range in 'integer->char'", m);
	return mkchar(c);
end

! Evaluate N=(INTEGER->LIST M)
dointegertolist(n) do var m;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	m := Car[m];
	if (atomic(m) \/ m = NIL \/ Car[m] \= S_number)
		return error("non-integer in 'integer->list'", m);
	return Cdr[m];
end

! Evaluate N=(LIST->INTEGER M)
dolisttointeger(n) do var m, check, p, d;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL /\ Cdr[Cdr[m]] \= NIL)
		return wrongargs(n);
	check := Cdr[Cdr[m]] \= NIL;
	m := Car[m];
	if (atomic(m) \/ tagged(m))
		return error("non-list in 'list->integer'", m);
	if (m = NIL)
		return error("empty list in 'list->integer'", m);
	if (check) do
		p := m;
		d := Car[p];
		if (	atomic(d) /\
			(Car[Car[d]] & ~AFLAG = '+' \/
			 Car[Car[d]] & ~AFLAG = '-') /\
			Cdr[Car[d]] = NIL
		)
			p := Cdr[p];
		while (p \= NIL) do
			d := Car[p];
			if (	d \= S_0 /\ d \= S_1 /\ d \= S_2 /\
				d \= S_3 /\ d \= S_4 /\ d \= S_5 /\
				d \= S_6 /\ d \= S_7 /\ d \= S_8 /\
				d \= S_9
			)
				error(
				"non-digit in argument to 'list->integer'",
					d);

			p := Cdr[p];
		end
	end
	return alloc(S_number, m);
end

! Evaluate N=(NULL M)
donullp(n) do var m;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	return Car[m] = NIL-> S_true: S_false;
end

! Evaluate N=(EQ? M M2)
doeqp(n) do var m;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] = NIL \/ Cdr[Cdr[m]] \= NIL)
		return wrongargs(n);
	return Car[m] = Car[Cdr[m]]-> S_true: S_false;
end


! Evaluate N=(LIST->STRING M) 
dolisttostring(n) do var m, a, ch, p;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	m := Car[m];
	if (m = NIL) do
		p := alloc(S_string, NIL);
		save(p);
		Cdr[p] := alloc(NIL, NIL);
		unsave(1);
		return p;
	end
	if (lazyatom(m)) return error("non-list in 'list->string'", m);
	p := alloc(NIL, NIL);
	save(p);
	a := p;
	while (m \= NIL) do
		ch := Car[m];
		if (atomic(ch) \/ Car[ch] \= S_char)
			return error("non-char in argument to 'list->string'",
				ch);
		Car[a] := Car[Cdr[ch]];
		m := Cdr[m];
		if (m \= NIL) do
			Cdr[a] := alloc(NIL, NIL);
			a := Cdr[a];
		end
	end
	p := alloc(p, NIL);
	Car[Stack] := p;
	p := alloc(S_string, p);
	unsave(1);
	return p;
end

! Evaluate N=(NUMBER? M)
donumberp(n) do var m;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	m := Car[m];
	return atomic(m) \/ m = NIL-> S_false:
		Car[m] = S_number-> S_true: S_false;
end

! Evaluate N=(PAIR? M)
dopairp(n) do var m;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	m := Car[m];
	return atomic(m) \/ tagged(m) \/ m = NIL-> S_false:
		S_true;
end

! Evaluate N=(PROCEDURE? M)
doprocedurep(n) do var m;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	m := Car[m];
	if (atomic(m) \/ m = NIL)
		return S_false;
	if (	Car[m] = S_closure \/
		Car[m] = S_continuation \/
		Car[m] = S_primitive
	)
		return S_true;
	return S_false;
end

! Evaluate N=(STRING->SYMBOL M) 
dostringtolist(n) do var m, a, p, lst;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	m := Car[m];
	if (atomic(m) \/ m = NIL \/ Car[m] \= S_string)
		return error("non-string in 'string->list'", m);
	p := Car[Cdr[m]];
	if (p = NIL) return NIL;
	lst := alloc(NIL, NIL);
	save(lst);
	a := lst;
	while (p \= NIL) do
		Car[a] := alloc(S_char, NIL);
		Cdr[Car[a]] := alloc(Car[p], NIL);
		p := Cdr[p];
		if (p \= NIL) do
			Cdr[a] := alloc(NIL, NIL);
			a := Cdr[a];
		end
	end
	unsave(1);
	return lst;
end

! Evaluate N=(STRING->SYMBOL M) 
dostringtosymbol(n) do var m, s::TEXTLEN, i, p;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	m := Car[m];
	if (atomic(m) \/ m = NIL \/ Car[m] \= S_string)
		return error("non-string in 'string->symbol'", m);
	if (Cdr[m] = NIL \/ Car[Cdr[m]] = NIL)
		return error("empty string in 'string->symbol'", m);
	p := Car[Cdr[m]];
	i := 0;
	while (p \= NIL) do
		if (i >= TEXTLEN-2)
			fatal("symbol too long in 'string->symbol'");
		s::i := Car[p] & ~AFLAG;
		i := i+1;
		p := Cdr[p];
	end
	s::i := 0;
	return addsym(s, S_void);
end

! Evaluate N=(STRING? M)
dostringp(n) do var m;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	m := Car[m];
	return \atomic(m) /\ Car[m] = S_string -> S_true: S_false;
end

! Evaluate N=(SYMBOL->STRING M) 
dosymboltostring(n) do var m, p;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	m := Car[m];
	if (\atomic(m))
		return error("non-symbol in 'symbol->string'", m);
	p := alloc(Car[m], NIL);
	save(p);
	p := alloc(S_string, p);
	unsave(1);
	return p;
end


! Evaluate N=(SYMBOL? M) 
dosymbolp(n) do var m, p;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	m := Car[m];
	return atomic(m) /\ m \= S_true /\ m \= S_false ->
		S_true: S_false;
end

! Extract clause from argument list of COND.
! Check the syntax of the clause.
! Return the predicate of the clause for
! evaluation in EVAL.
getpred() do var e;
	e := Car[Car[Bstack]];
	if (	atomic(e) \/ e = NIL \/
		Cdr[e] = NIL \/ Cdr[Cdr[e]] \= NIL
	)
		return error("bad clause in 'cond'", e);
	return Car[e];	! predicate
end

! Setup context for evaluation of (COND (P1 E1) ... (Pn En)).
! The context consits of a list of clauses.
! return the predicate of the first clause.
setupcond(n) do var m;
	m := Cdr[n];
	if (m = NIL) return wrongargs(n);
	bsave(m);
	return getpred();
end

! Evaluate next clause of COND.
! N is the value of the current predicate.
! If N=*T*, return the expression of the predicate.
! If N=*F*, return the predicate of the next clause.
! When returning the expression of a predicate (N=*T*),
! set the context on the Bstack to NIL to signal that
! a true clause was found.
! Clauses and predicates are returned to EVAL for
! evaluation.
evalclause(n) do var e;
	e := Car[Bstack];
	ie (n = S_false) do
		Car[Bstack] := Cdr[e];
		if (Car[Bstack] = NIL)
			return error("no default in 'cond'", %1);
		return getpred();
	end
	else ie (n = S_true) do
		e := Car[Cdr[Car[e]]];
		Car[Bstack] := NIL;
		return e;
	end
	else do
		return error("bad predicate in 'cond'", Car[Car[e]]);
	end
end

! Unbind the arguments of LAMBDA and LETREC.
! See also BINDARGS().
unbindargs() do var v;
	Frame := unsave(1);
	Function := unsave(1);
	v := bunsave(1);		! Caller's namelist
	while (v \= NIL) do
		Cdr[Car[v]] := unsave(1);
		v := Cdr[v];
	end
end

! Check if the symbol N is bound in the current lexical
! environment.
isbound(n) do var b;
	b := Bound;
	while (b \= NIL) do
		if (atomic(b)) return n = b;
		if (n = Car[b]) return 1;
		b := Cdr[b];
	end
	b := Car[Lexenv];
	while (b \= NIL) do
		if (Car[Car[b]] = n) return 1;
		b :=Cdr[b];
	end
	return 0;
end

! Recursively collect free variables and add their symbols
! and values to the current lexical environment.
collect(n) do
	if (n = NIL \/ n & AFLAG \/ Car[n] & AFLAG) return;
	if (Car[Car[n]] & AFLAG) do
		if (isbound(n)) return;
		Car[Lexenv] := alloc(NIL, Car[Lexenv]);
		Car[Car[Lexenv]] := alloc(n, Car[n] = Cdr[n]-> n: Cdr[n]);
		return;
	end
	collect(Car[n]);
	collect(Cdr[n]);
end

! Create lexical environment.
mklexenv(term, locals) do var env;
	Lexenv := alloc(NIL, NIL);
	save(Lexenv);
	Bound := atomic(locals)-> alloc(locals, NIL) : locals;
	collect(term);
	unsave(1);
	return Car[Lexenv];
end

! Create a closure from a lambda expression.
closure(n) do var cl, env, args, term;
	if (Errflag) return NIL;
	cl := alloc(S_closure, NIL);
	save(cl);
	args := Car[Cdr[n]];
	Cdr[cl] := alloc(args, NIL);
	term := Car[Cdr[Cdr[n]]];
	Cdr[Cdr[cl]] := alloc(term, NIL);
	ie (Cdr[Cdr[Cdr[n]]] = NIL) do
		env := mklexenv(term, args);
		save(env);
		if (env \= NIL) do
			Cdr[Cdr[Cdr[cl]]] := alloc(env, NIL);
			Estack := alloc(env, Estack);
		end
		unsave(1);
	end
	else do
		Cdr[Cdr[Cdr[cl]]] := alloc(Car[Cdr[Cdr[Cdr[n]]]], NIL);
	end
	unsave(1);
	return cl;
end

! Fix recursive bindings in closures.
fixclosures() do var aa, a, ee, e;
	if (Errflag \/ Estack = NIL) return;
	a := Car[Bstack];
	while (a \= NIL) do
		ee := Estack;
		while (ee \= NIL) do
			e := Car[ee];
			while (e \= NIL) do
				if (Car[a] = Car[Car[e]]) do
					Cdr[Car[e]] := Cdr[Car[a]];
					leave;
				end
				e := Cdr[e];
			end
			ee := Cdr[ee];
		end
		a := Cdr[a];
	end
	Estack := NIL;
end

! Set up a context for processing N=(LETREC ((MA1 eval[MX2]) ...) MN)
! Save
! - the complete LETREC expression on the Bstack
! - the environment on the Bstack
! - a list of new bindings on the Bstack (initially empty)
! - two copies of the saved name list on the Stack
setupletrec(n) do
	var	m,	! Argument pointer
		v;	! List of re-bound names

	m := Cdr[n];
	if (m = NIL \/ Cdr[m] = NIL \/ Cdr[Cdr[m]] \= NIL)
		return wrongargs(n);
	m := Car[m];
	if (atomic(m)) return error("bad environment in 'letrec'", m);
	bsave(n);	! save entire LETREC
	bsave(m);	! save environment
	bsave(NIL);	! list of bindings
	bsave(NIL);	! save empty name list
	return m;
end

! Process one binding of a LETREC:
! bind value to name, advance to next binding.
! Return:
! non-NIL - more bindings in environment
! NIL     - last binding done
nextletrec(n) do var m, p, r;
	m := Car[Cdr[Cdr[Bstack]]];	! rest of environment
	if (m = NIL) return NIL;
	p := Car[m];
	Tmp2 := n;
	Car[Cdr[Bstack]] := alloc(NIL, Car[Cdr[Bstack]]);
	Car[Car[Cdr[Bstack]]] := alloc(Car[p], n);
	Tmp2 := NIL;
	Car[Cdr[Cdr[Bstack]]] := Cdr[m];
	return Cdr[m];
end

! Evaluate value to bind inside a LETREC:
! - check LETREC syntax
! - save name to bind to
! - save original binding of name
! - return value for evaluation in EVAL
evalletrec() do
	var	m, p, v;

	m := Car[Cdr[Cdr[Bstack]]];
	p := Car[m];
	! Each binding must have the form (atom expr)
	if (	atomic(p) \/ Cdr[p] = NIL \/ atomic(Cdr[p]) \/
		Cdr[Cdr[p]] \= NIL \/ \atomic(Car[p])
	) do
		! In case of an error, get rid of the
		! partial environment.
		v := bunsave(1);
		bunsave(3);
		bsave(v);
		save(Function);
		save(Frame);
		unbindargs();
		return error("bad binding in 'letrec'", p);
	end
	Car[Bstack] := alloc(Car[p], Car[Bstack]);	! Save name
	! Evaluate the new value of the current symbol
	return Car[Cdr[p]];
end

! Establish the bindings of LETREC.
bindletrec(env) do var e, b;
	e := env;
	while (e \= NIL) do
		b := Car[e];
		save(Cdr[Car[b]]);	! Save old value
		Cdr[Car[b]] := Cdr[b];	! Bind new value
		e := Cdr[e];
	end
end

decl dumpstate(2);

! Finish processing bindings of LETREC:
! finish context and return term of LETREC.
finishletrec() do var m, v, b;
	Tmp2 := alloc(NIL, NIL);	! Create safe storage
	Cdr[Tmp2] := alloc(NIL, NIL);
	Cdr[Cdr[Tmp2]] := alloc(NIL, NIL);
	v := bunsave(1);
	b := bunsave(1);	! get bindings
	m := bunsave(2);	! drop environment, get full LETREC
	Car[Tmp2] := b;		! protect b, m, v
	Car[Cdr[Tmp2]] := m;
	Cdr[Cdr[Tmp2]] := v;
	bindletrec(b);
	bsave(v);
	fixclosures();
	save(Function);			! required by UNBINDARGS()
	save(Frame);
	Tmp2 := NIL;
	return Car[Cdr[Cdr[m]]];	! return term of LETREC
end

decl eval(1);

! Evaluate N=(DEFINE M eval[M2])
! The name M already has been added to the
! symbol table by READ().
dodefine(n) do var m, v;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] = NIL \/ Cdr[Cdr[m]] \= NIL)
		return wrongargs(n);
	if (\atomic(Car[m])) return error("bad symbol name", Car[m]);
	! Bind (and thereby protect) the unevaluated expression
	! Cdr[Car[m]] := Car[Car[m]];
	! Protect the unevaluated expression
	v := Car[Cdr[m]];
	save(v);
	if (\atomic(v) /\ Car[v] = S_lambda) do
		if (	Cdr[v] \= NIL /\ Cdr[Cdr[v]] \= NIL /\
			Cdr[Cdr[Cdr[v]]] = NIL
		) do
			Cdr[Cdr[Cdr[v]]] := alloc(NIL, NIL);
		end
	end
	! And then evaluate it
	Cdr[Car[m]] := eval(Car[Cdr[m]]);
	unsave(1);
	return Car[m];
end

! Evaluate N=(QUOTE M)
doquote(n) do var m;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	return (Car[m]);
end

! Evaluate N=(APPLY M)
doapply(n) do var m, p;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] = NIL \/ Cdr[Cdr[m]] \= NIL)
		return wrongargs(n);
	p := Car[Cdr[m]];
	while (p \= NIL) do
		if (atomic(p) \/ tagged(p)) return
			error("improper list in 'apply'", Car[Cdr[m]]);
		p := Cdr[p];
	end
	return alloc(Car[m], Car[Cdr[m]]);
end

! Check if N is an alist.
isalist(n) do
	if (atomic(n)) return 0;
	while (n \= NIL) do
		if (lazyatom(Car[n]) \/ \atomic(Car[Car[n]]))
			return 0;
		n := Cdr[n];
	end
	return %1;
end

! Evaluate N=(BOTTOM ...)
dobottom(n) return error("", n);

! Check LAMBDA syntax and create closure from lambda expression.
dolambda(n) do var m, a;
	m := Cdr[n];
	if (	m = NIL \/ Cdr[m] = NIL \/
		Cdr[Cdr[m]] \= NIL /\ Cdr[Cdr[Cdr[m]]] \= NIL
	)
		return wrongargs(n);
	if (Cdr[Cdr[m]] \= NIL /\ \isalist(Car[Cdr[Cdr[m]]]))
		return error("bad environment in 'lambda'",
			Car[Cdr[Cdr[m]]]);
	if (\atomic(Car[m])) do
		if (tagged(Car[m])) return badargl(Car[m]);
		m := Car[m];
		a := m;
		while (m \= NIL) do
			if (\atomic(Car[m])) return badargl(a);
			if (atomic(Cdr[m])) leave;
			m := Cdr[m];
		end
	end
	return Car[n] = S_closure-> n: closure(n);
end

! Create a flat copy of a list.
! Store a reference to the last member
! of the copy in lastp.
flatcopy(n, lastp) do var a, m, last;
	if (n = NIL) do
		lastp[0] := NIL;
		return NIL;
	end
	m := alloc(NIL, NIL);
	save(m);
	a := m;
	last := m;
	while (n \= NIL) do
		Car[a] := Car[n];
		last := a;
		n := Cdr[n];
		if (n \= NIL) do
			Cdr[a] := alloc(NIL, NIL);
			a := Cdr[a];
		end
	end
	unsave(1);
	lastp[0] := last;
	return m;
end

! Create a 2-level copy of the Lstack
! and fix cyclic references. N = Lstack.
copy(n) do var m, a, p, last, unused;
	if (n = NIL) return NIL;
	m := alloc(NIL, NIL);
	a := m;
	save(m);
	while (n \= NIL) do
		if (Cdr[n] = NIL \/ Cdr[Cdr[n]] = NIL)
			fatal("internal: unexpected NIL in COPY");
		Car[a] := flatcopy(Car[n], @last);
		Cdr[a] := alloc(NIL, NIL);
		Car[Cdr[a]] := flatcopy(Car[Cdr[n]], @unused);
		Cdr[Cdr[a]] := alloc(last, NIL);
		n := Cdr[Cdr[Cdr[n]]];
	end
	unsave(1);
	return m;
end

! Copy names and values of the symbol table
! into an alist.
copybindings() do var y, p, ny;
	y := Symbols;
	p := alloc(NIL, NIL);
	ny := p;
	save(p);
	while (y \= NIL) do
		Car[p] := alloc(Car[y], Cdr[Car[y]]);
		y := Cdr[y];
		if (y \= NIL) Cdr[p] := alloc(NIL, NIL);
		p := Cdr[p];
	end
	unsave(1);
	return ny;
end

! Restore values of the symbol table.
restorebindings(values) do var b;
	while (values \= NIL) do
		b := Car[values];
		Cdr[Car[b]] := Cdr[b];
		values := Cdr[values];
	end
end

! Capture interpreter state, creating a continuation.
capture() do var cont, x;
	cont := alloc(S_continuation, NIL);
	save(cont);
	Cdr[cont] := alloc(NIL, NIL);
	Car[Cdr[cont]] := alloc(Car[Mstack], NIL);
	Cdr[Cdr[cont]] := alloc(NIL, NIL);
	Car[Cdr[Cdr[cont]]] := copy(Lstack);
	Cdr[Cdr[Cdr[cont]]] := alloc(Cdr[Cdr[Stack]], NIL);
	Cdr[Cdr[Cdr[Cdr[cont]]]] := alloc(NIL, NIL);
	Car[Cdr[Cdr[Cdr[Cdr[cont]]]]] := flatcopy(Bstack, @x);
	Cdr[Cdr[Cdr[Cdr[Cdr[cont]]]]] := alloc(Estack, NIL);
	Cdr[Cdr[Cdr[Cdr[Cdr[Cdr[cont]]]]]] := alloc(NIL, NIL);
	Car[Cdr[Cdr[Cdr[Cdr[Cdr[Cdr[cont]]]]]]] := copybindings();
	unsave(1);
	return cont;
end

! Check the arguments of a continuation.
ckcont(n) do
	if (	Cdr[n] = NIL \/ Cdr[Cdr[n]] = NIL \/
		Cdr[Cdr[Cdr[n]]] = NIL \/ Cdr[Cdr[Cdr[Cdr[n]]]] = NIL \/
		Cdr[Cdr[Cdr[Cdr[Cdr[n]]]]] = NIL \/
		Cdr[Cdr[Cdr[Cdr[Cdr[Cdr[n]]]]]] = NIL \/
		Cdr[Cdr[Cdr[Cdr[Cdr[Cdr[Cdr[n]]]]]]] \= NIL
	)
		error("invalid continuation", %1);
	return n;
end

! Re-activate a captured continuation N.
resume(n) do var m, state, x;
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	ckcont(Car[n]);
	if (Errflag) return NIL;
	state := Cdr[Car[n]];
	Car[Mstack] := Car[Car[state]];
	Lstack := Car[Cdr[state]];
	Stack := Car[Cdr[Cdr[state]]];
	Bstack := flatcopy(Car[Cdr[Cdr[Cdr[state]]]], @x);
	Estack := Car[Cdr[Cdr[Cdr[Cdr[state]]]]];
	restorebindings(Car[Cdr[Cdr[Cdr[Cdr[Cdr[state]]]]]]);
	return Car[m];
end

! Evaluate N=(CALL/CC M)
docallcc(n) do var m;
	if (Evlev > 1) do
		error("'call/cc' cannot be used here; this is a known bug",
		%1);
		return NIL;
	end
	m := Cdr[n];
	if (m = NIL \/ Cdr[m] \= NIL) return wrongargs(n);
	m := alloc(Car[m], NIL);
	save(m);
	Cdr[m] := alloc(capture(), NIL);
	unsave(1);
	return m;
end

! Check whether (CAR NP[0]) is a builtin symbol.
! If it is one, run the appropriate routine, save
! its result in NP[0], and return %1.
! Return 0 if (CAR NP[0]) is not a builtin procedure.
primitive(np) do var n, y, op;
	n := np[0];
	y := Car[n];
	if (Car[y] \= S_primitive \/ Errflag) return 0;
	op := Primitives[Car[Cdr[y]] & ~AFLAG];
	n := call op(n);
	np[0] := n;
	return %1;
end

! Bind the arguments of a LAMBDA function.
! For a lambda application N=((LAMBDA (X1 ... Xn) S [ENV]) Y1 ... Yn)
! this includes the following steps for j in {1,...,n}:
!	1) save Xj in V
!	2) save the value of Xj
!	3) bind Xj to Yj
! This routine results in  S' = S[X1/Y1] ... [Xn/Yn].
! S->S' is performed by creating a new context. BINDARGS()
! has no function result. It is used by EVAL() to perform
! BETA reduction steps.
bindargs(n, name) do
	var	fa,	! Formal arg list
		aa,	! Actual arg list
		e;	! S, as above
	var	env;	! Optional lexical environment
	var	p;
	var	at;	! Atomic argument list flag

	if (Errflag) return NIL;
	fa := Car[Cdr[Car[n]]];
	at := atomic(fa);
	aa := Cdr[n];
	p := Cdr[Cdr[Car[n]]];
	e := Car[p];
	env := Cdr[p] \= NIL -> Car[Cdr[p]]: NIL;
	bsave(NIL);
	while (fa \= NIL /\ aa \= NIL \/ at) do
		if (\at) do
			Car[Bstack] := alloc(Car[fa], Car[Bstack]); ! Save name
			save(Cdr[Car[fa]]);			! Save value
			Cdr[Car[fa]] := Car[aa];		! Bind arg
			fa := Cdr[fa];
			aa := Cdr[aa];
		end
		if (atomic(fa)) do	! improper argument list
			Car[Bstack] := alloc(fa, Car[Bstack]);	! Save name
			save(Cdr[fa]);				! Save value
			Cdr[fa] := aa;	! Bind remaining arg list
			fa := NIL;
			aa := NIL;
			leave;
		end
	end
	while (env \= NIL) do
		p := Car[env];
		Car[Bstack] := alloc(Car[p], Car[Bstack]);! Save name
		save(Cdr[Car[p]]);		! Save value
		Cdr[Car[p]] := Cdr[p];		! Bind lex val
		env := Cdr[env];
	end
	ie (fa \= NIL \/ aa \= NIL) do
		wrongargs(n);
		n := NIL;
	end
	else do
		n := e;
	end
	save(Function);
	Function := name;
	save(Frame);
	Frame := Stack;
	return n;
end

! Print application of traced function N in the form
!	* (NAME A1 ... An)
! PRINT() cannot be used because it would print NAME in
! its expanded (LAMBDA...) form which is not desirable.
printtrace(n) do
	Prefix := "+ ";
	pr("+ ");
	pr("(");
	Quoted := 1;
	_print(Trace);
	while (1) do
		n := Cdr[n];
		if (n = NIL) leave;
		pr("\s");
		_print(Car[n]);
	end
	pr(")"); nl();
	Prefix := 0;
end

! Print depth of stack N.
prdepth(n) do var k;
	k := 0;
	while (n \= NIL) do
		n := Cdr[n];
		k := k+1;
	end
	prnum(k, 6); pr("\s");
end

! Debugging: dump interpreter state
dumpstate(s, m) do
	pr(s); nl();
	pr("\s\s\s\s\s\s\s"); pr("Mode        = "); prnum(m-'0',0); nl();
	prdepth(Lstack);      pr("Lstack      = "); print(Lstack); nl();
	pr("\s\s\s\s\s\s\s"); pr("Mstack      = "); print(Mstack); nl();
	prdepth(Stack);       pr("Car[Stack]  = "); print(Car[Stack]); nl();
	pr("\s\s\s\s\s\s\s"); pr("Stack0      = "); prnum(Stack0,0); nl();
	prdepth(Bstack);      pr("Bstack      = "); print(Bstack); nl();
end

! Evaluator states
const	MATOM = '0',	! Processing Atom
	MLIST = '1',	! Processing List
	MBETA = '2',	! Beta-reducing
	MBIND = '3',	! Processing bindings of LETREC
	MLETR = '4',	! Finish LETREC
	MCOND = '5';	! Processing predicates of COND

! Do tail call optimization.
tailcall() do var m, y;
	m := Car[Mstack];
	! Skip over callee's LETREC frames, if any
	while (m \= NIL /\ Car[m] & ~AFLAG = MLETR) do
		m := Cdr[m];
	end
	! Parent not beta-reducing? Give up.
	if (m = NIL \/ Car[m] & ~AFLAG \= MBETA)
		return;
	! Yes, this is a tail call:
	! - remove callee's LETREC frames.
	! - remove caller's call frame.
	while (1) do
		Tmp2 := unsave(1); ! M
		unbindargs();
		unsave(1);
		y := munsave();
		save(Tmp2);
		Tmp2 := NIL;
		if (y = MBETA) leave;
	end
end

! Evaluate the term N and return its normal form.
! This is the heart of the interpreter:
! An iterative EVAL function with tail-call optimization and call/cc.
! For details on the reduction rules, refer to the manual.
! Basically, EVAL() performs these steps:
!	- If N is NIL, return NIL
!	- If N is a symbol, return its binding
!	- If N is a list (A1 ... An)
!		- If A1 not in {COND,DEFINE,LETREC,LAMBDA,QUOTE}
!			- Create a new list (A1' ... An')
!				where Aj'=EVAL(Aj) for j in {1,...,n}
!		- If A1' (or A1) is a primitive (builtin fn), apply it,
!			resulting in N'
!		- If A1 is a lambda function, BETA-reduce it,
!			resulting in N'
!	- Return N'
!
eval(n) do
	var	m,	! Result node
		m2,	! Root of result lists
		a,	! Used to append to result
		cbn;	! Use call-by-name/quotation in next iteration
	var	mode,	! Current state
		cf;	! Continue flag
	var	nm;	! name of function to apply

	Evlev := Evlev + 1;
	save(Stack0);
	Stack0 := Stack;
	mode := MATOM;
	cf := 0;
	cbn := 0;
	while (\Errflag) do
		if (Statflag) Reductions.count(1);
		ie (n = NIL) do			! () -> ()
			m := NIL;
		end
		else ie (atomic(n)) do		! Symbol -> Value
			ie (cbn) do
				m := n;
				cbn := 0;
			end
			else do
				m := Cdr[n] = Car[n]-> n: Cdr[n];
				if (m = S_void)
					return error("symbol not bound", n);
			end
		end
		else ie (Car[n] = S_closure \/
			Car[n] = S_primitive \/
			Car[n] = S_continuation \/
			Car[n] = S_number \/
			Car[n] = S_char \/
			Car[n] = S_string \/
			cbn = 2
		) do
			m := n;
			cbn := 0;
		end
		else do				! List (...) and Pair (X.Y)
			! This block is used to DESCEND into lists.
			! The following nodes/variables will be saved:
			!	1) the original list (on Stack)
			!	2) the current state (on Mstack)
			!	3) the root of the result list (on Lstack)
			!	4) a ptr to the next free node
			!		in the result list (on Lstack)
			!	5) a ptr to the next member of
			!		the original list (on Lstack)
			m := Car[n];
			if (atomic(Cdr[n])) do
				error("improper list in application", n);
				n := NIL;
			end
			save(n);	! Save original list
			msave(mode);
			! Check call-by-name built-ins and flag
			ie (	m = S_cond \/ m = S_define \/
				m = S_letrec \/ m = S_quote \/
				m = S_closure \/ m = S_continuation \/
				m = S_lambda \/ cbn
			) do
				if (cbn) cbn := 0;
				lsave(NIL);
				lsave(NIL);
				lsave(n);	! Root of result list
				n := NIL;
			end
			else do
				a := alloc(NIL, NIL);
				lsave(a);
				lsave(Cdr[n]);
				lsave(a);	! Root of result list
				n := Car[n];
			end
			mode := MLIST;
			loop;
		end
		! The following loop is used to ASCEND back to the
		! root of a list, thereby performing BETA reduction
		! and creating result lists.
		while (1) ie (mode = MBETA \/ mode = MLETR) do
			! Finish BETA reduction
			unbindargs();
			unsave(1);	! Original list
			mode := munsave();
		end
		else ie (mode = MLIST) do	! Append to list, reduce
			n := Car[Cdr[Lstack]];	! Next member
			a := Car[Cdr[Cdr[Lstack]]];	! Place to append to
			m2 := lunsave(1);	! Root of result list
			if (a \= NIL)		! Append new member
				Car[a] := m;
			ie (n = NIL) do		! End of list
				m := m2;
				lunsave(2);	! Drop N,A
 				! Drop original list, remember first element
				nm := Car[unsave(1)];
				save(m);	! Save result
				if (Trace = nm) printtrace(m);
				ie (primitive(@m))	! primitive fn?
					;
				else ie (Car[m] = S_quote) do
					! QUOTE is special
					m := doquote(m);
				end
				else ie (Car[m] = S_cond) do
					! COND is special
					n := setupcond(m);
					cf := 2;
					mode := MCOND;
				end
				else ie (\atomic(Car[m]) /\
					Car[m] \= NIL /\
					Car[Car[m]] = S_closure
				) do
					! Application: reduce ((lambda...)...)
					nm := atomic(nm)-> nm: NIL;
					! If the caller is also BETA-reducing,
					! this is a TAIL application.
					tailcall();
					bindargs(m, nm);
					! N=S of ((LAMBDA (...) S) ...)
					n := Car[Cdr[Cdr[Car[m]]]];
					cf := 2;
					mode := MBETA;
				end
				else ie (Car[m] = S_lambda) do
					! LAMBDA is special
					m := dolambda(m);
				end
				else ie (Car[m] = S_letrec) do
					! LETREC is special
					ie (setupletrec(m) \= NIL)
						n := evalletrec();
					else
						n := NIL;
					cf := 2;
					mode := MBIND;
				end
				else ie (Car[m] = S_callcc) do
					! CALL/CC is special
					n := docallcc(m);
					cf := 1;
				end
				else ie (\atomic(Car[m]) /\
					Car[m] \= NIL /\
					Car[Car[m]] = S_continuation
				) do
					n := resume(m);
					cbn := 2;
					cf := 1;
				end
				else ie (Car[m] = S_apply) do
					! APPLY is special
					n := doapply(m);
					cbn := 1;
					cf := 1;
				end
				else ie (Car[m] = S_define) do
					! DEFINE is special
					m := dodefine(m);
				end
				else do
					error("application of non-function",
						nm);
					n := NIL;
				end
				if (cf \= 2) do
					unsave(1);	! Drop result
					mode := munsave();
				end
				! Continue this evaluation.
				! Leave the ASCENDING loop and descend
				! once more into N.
				if (cf) leave;
			end
			else do			! Append to list
				lsave(m2);
				! Create space for next argument
				Cdr[a] := alloc(NIL, NIL);
				Car[Cdr[Cdr[Lstack]]] := Cdr[a];
				Car[Cdr[Lstack]] := Cdr[n];
				n := Car[n];	! Evaluate next member
				leave;
			end
		end
		else ie (mode = MCOND) do
			n := evalclause(m);
			if (Car[Bstack] = NIL) do
				unsave(1);	! Original list
				bunsave(1);
				mode := munsave();
			end
			cf := 1;
			leave;
		end
		else ie (mode = MBIND) do
			ie (nextletrec(m) = NIL) do
				n := finishletrec();
				mode := MLETR;
			end
			else do
				n := evalletrec();
			end
			cf := 1;
			leave;
		end
		else do	! Atom
			leave;
		end
		if (cf) do	! Continue evaluation if requested
			cf := 0;
			loop;
		end
		if (Stack = Stack0) leave;
	end
	while (Stack \= Stack0) unsave(1);
	Stack0 := unsave(1);
	Evlev := Evlev - 1;
	return m;		! Return the evaluated expr
end

! Print lists of digits in condensed format.
printnum(n) do var s::2, c;
	if (Car[n] \= S_number) return 0;
	s::1 := 0;
	n := Cdr[n];
	while (1) do
		if (n = NIL) leave;
		s::0 := Car[Car[Car[n]]] & 255;
		pr(s);
		n := Cdr[n];
	end
	return %1;
end

! Print expressions of the form (QUOTE X) as 'X.
printquote(n) do
	if (	Car[n] = S_quote /\
		Cdr[n] \= NIL /\
		Cdr[Cdr[n]] = NIL
	) do
		n := Car[Cdr[n]];
		if (n \= S_true /\ n \= S_false) pr("'");
		_print(n);
		return 1;
	end
	return 0;
end

! Print a closure.
printclosure(n) do
	if (	Car[n] = S_closure /\
		Cdr[n] \= NIL /\ \atomic(Cdr[n]) /\
		Cdr[Cdr[n]] \= NIL /\ \atomic(Cdr[Cdr[n]])
	) do
		Quoted := 1;
		pr("#<closure ");
		_print(Car[Cdr[n]]);
		if (Prcllevl > 0) do
			pr("\s");
			_print(Car[Cdr[Cdr[n]]]);
			if (Prcllevl > 1 /\ Cdr[Cdr[Cdr[n]]] \= NIL) do
				pr("\s");
				_print(Car[Cdr[Cdr[Cdr[n]]]]);
			end
		end
		pr(">");
		return -1;
	end
	return 0;
end

! Print a continuation.
printcont(n)
	ie (Car[n] = S_continuation) do
		pr("#<continuation>");
		return %1;
	end
	else do
		return 0;
	end

! Print a character.
printchar(n) do var b::2, c;
	if (Car[n] \= S_char) return 0;
	pr("#\\");
	c := Car[Cdr[n]] & ~AFLAG;
	ie (c = '\s') do
		pr("space");
	end
	else do
		b::1 := 0;
		b::0 := c;
		pr(b);
	end
	return %1;
end

! Print a string.
printstring(n) do var b::2, c, p;
	if (Car[n] \= S_string) return 0;
	pr("\q");
	p := Car[Cdr[n]];
	b::1 := 0;
	while (p \= NIL) do
		b::0 := Car[p] & ~AFLAG;
		if (b::0 = '"' \/ b::0 = '\\') pr("\\");
		pr(b);
		p := Cdr[p];
	end
	pr("\q");
	return %1;
end

! Print a primitive function.
printprim(n) do
	if (Car[n] \= S_primitive) return 0;
	_print(Cdr[Cdr[n]]);
	return %1;
end

! Recursively print the term N.
_print(n) do
	ie (n = NIL) do
		if (\Quoted) do
			pr("'");
			Quoted := 1;
		end
		pr("()");
	end
	else ie (Car[n] & AFLAG) do
		! Characters are limited to the symbol table
		pr("#<unprintable object>");
	end
	else ie (atomic(n)) do var s::TEXTLEN, i;
		if (\Quoted) do
			if (	n \= S_true /\
				n \= S_false /\
				n \= S_number
			)
				pr("'");
			Quoted := 1;
		end
		i := 0;		! Symbol
		n := Car[n];
		while (n \= NIL) do
			s::i := Car[n];
			if (i < TEXTLEN-2) i := i+1;
			n := Cdr[n];
		end
		s::i := 0;
		pr(s);
	end
	else do	! List
		if (printnum(n)) return;
		if (printchar(n)) return;
		if (printstring(n)) return;
		if (printcont(n)) return;
		if (printclosure(n)) return;
		if (printprim(n)) return;
		if (\Quoted) do
			pr("'");
			Quoted := 1;
		end
		if (printquote(n)) return;
		pr("(");
		while (n \= NIL) do
			_print(Car[n]);
			n := Cdr[n];
			if (	n \= NIL /\
				(atomic(n) \/
				 Car[n] = S_number \/
				 Car[n] = S_primitive)
			) do
				pr(".");
				_print(n);
				n := NIL;
			end
			if (n \= NIL) pr("\s");
		end
		pr(")");
	end
end

! Print quoted expression.
print(n) do
	Quoted := 0;
	_print(n);
end

! Reset interpreter state.
resetstate() do
	Stack := NIL;
	Lstack := NIL;
	Bstack := NIL;
	Estack := NIL;
	Frame := NIL;
	Function := NIL;
	Evlev := 0;
end

! Initialize interpreter variables.
init1() do var i, n;
	RELEASE := "2005-07-05";
	! Misc. variables
	NIL := POOLSIZE;
	Level := 0;
	resetstate();
	Symbols := NIL;
	Safesymbols := NIL;
	Tmp := NIL;
	Tmp2 := NIL;
	Loadlev := 0;
	Trace := NIL;
	Pp := 0;
	t.newline(Nlseq);
	Nll := str.length(Nlseq);
	Wrapcol := 0;
	Prefix := 0;
	Maxuse := 0;
	Lastuse := 0;
	Statflag := 0;
	Ntrace := 10;
	Prcllevl := 0;
	Image::0 := 0;
	Line := 1;
	! Initialize Freelist
	Free := NIL;
	! Clear input buffer
	Infile := 0;
	Buffer := Bufvec;
	Input := T3X.SYSIN;
	Bp := 0;
	Ep := 0;
	Rejected := EOF;
end

! Second stage of initialization:
! protect registers from GC,
! build the free list,
! create builtin symbols.
init2() do
	! Protect base registers
	Base[0] := @Symbols;
	Base[1] := @Stack;
	Base[2] := @Mstack;
	Base[3] := @Lstack;
	Base[4] := @Bstack;
	Base[5] := @Estack;
	Base[6] := @Tmp;
	Base[7] := @Tmp2;
	Base[8] := @Safesymbols;
	! Create builtin symbols
	S_0 := addsym("0-digit", 0);	! First GC will be triggered HERE
	S_1 := addsym("1-digit", 0);
	S_2 := addsym("2-digit", 0);
	S_3 := addsym("3-digit", 0);
	S_4 := addsym("4-digit", 0);
	S_5 := addsym("5-digit", 0);
	S_6 := addsym("6-digit", 0);
	S_7 := addsym("7-digit", 0);
	S_8 := addsym("8-digit", 0);
	S_9 := addsym("9-digit", 0);
	! Tags (especially #<primitive>) must be defined
	! before any primitives.
	S_char := addsym("#<char>", 0);
	S_closure := addsym("#<closure>", 0);
	S_continuation := addsym("#<continuation>", 0);
	S_number := addprim("#<number>", 0);
	S_primitive := addsym("#<primitive>", 0);
	S_string := addsym("#<string>", 0);
	S_void := addsym("#<void>", 0);
	S_apply := addsym("apply", 0);
	addprim("bottom", P_BOTTOM);
	S_callcc := addsym("call/cc", 0);
	addprim("car", P_CAR);
	addprim("cdr", P_CDR);
	addprim("char->integer", P_CHAR_TO_INTEGER);
	addprim("char?", P_CHARP);
	S_cond := addsym("cond", 0);
	addprim("cons", P_CONS);
	S_define := addsym("define", 0);
	addprim("eq?", P_EQP);
	S_false := addsym("#f", 0);
	addprim("integer->char", P_INTEGER_TO_CHAR);
	addprim("integer->list", P_INTEGER_TO_LIST);
	S_lambda := addsym("lambda", 0);
	S_letrec := addsym("letrec", 0);
	addprim("list->integer", P_LIST_TO_INTEGER);
	addprim("list->string", P_LIST_TO_STRING);
	addprim("null?", P_NULLP);
	addprim("number?", P_NUMBERP);
	addprim("pair?", P_PAIRP);
	addprim("procedure?", P_PROCEDUREP);
	S_quote := addsym("quote", 0);
	addprim("procedure?", P_PROCEDUREP);
	addprim("string->symbol", P_STRING_TO_SYMBOL);
	addprim("string->list", P_STRING_TO_LIST);
	addprim("string?", P_STRINGP);
	addprim("symbol->string", P_SYMBOL_TO_STRING);
	addprim("symbol?", P_SYMBOLP);
	S_true := addsym("#t", 0);
	S_last := addsym("**", 0);
	Mstack := alloc(NIL, NIL);
	Primitives[P_BOTTOM] := @dobottom;
	Primitives[P_CAR] := @docar;
	Primitives[P_CDR] := @docdr;
	Primitives[P_CHAR_TO_INTEGER] := @dochartointeger;
	Primitives[P_CHARP] := @docharp;
	Primitives[P_CONS] := @docons;
	Primitives[P_EQP] := @doeqp;
	Primitives[P_INTEGER_TO_CHAR] := @dointegertochar;
	Primitives[P_INTEGER_TO_LIST] := @dointegertolist;
	Primitives[P_LIST_TO_INTEGER] := @dolisttointeger;
	Primitives[P_LIST_TO_STRING] := @dolisttostring;
	Primitives[P_NULLP] := @donullp;
	Primitives[P_NUMBERP] := @donumberp;
	Primitives[P_PAIRP] := @dopairp;
	Primitives[P_PROCEDUREP] := @doprocedurep;
	Primitives[P_STRING_TO_LIST] := @dostringtolist;
	Primitives[P_STRING_TO_SYMBOL] := @dostringtosymbol;
	Primitives[P_STRINGP] := @dostringp;
	Primitives[P_SYMBOL_TO_STRING] := @dosymboltostring;
	Primitives[P_SYMBOLP] := @dosymbolp;
end

! Reset the reduction counter
clearstats() do
	Reductions.reset();
	Allocations.reset();
end

! Print the number of reductions done in last EVAL
printstats() do
	Reductions.printvalue("calls to eval");
	Allocations.printvalue("nodes allocated");
end

! This is the top-level Read-Eval-Print loop.
toploop() do var n, def;
	while (1) do
		Errflag := 0;
		n := read();
		if (n = EOF) leave;
		save(n);
		Safesymbols := copybindings();
		if (Statflag) clearstats();
		n := eval(Car[Stack]);
		unsave(1);
		ie (\Errflag) do
			Cdr[S_last] := n;
			if (\Loadlev) do
				if (\O_batch) pr("=> ");
				print(n); nl();
			end
			if (Stack \= NIL)
				fatal("internal: unbalanced stack");
			if (Statflag /\ \Loadlev) printstats();
		end
		else do
			restorebindings(Safesymbols);
		end
		resetstate();
		while (Car[Mstack] \= NIL) munsave();
	end
end

! Load initial node pool image
loadimage(p) do var fd, n, v, i, buf::6;
	fd := t.open(p, T3X.OREAD);
	if (fd < 0) do
		error("cannot open image", %1);
		pr("* "); pr(p); nl();
		halt 1;
	end
	t.read(fd, buf, 9);
	t.read(fd, @n, t.bpw());
	if (	t.memcomp(buf, "SKETCHY", 7) \/
		buf::7 \= t.bpw() \/
		buf::8 \= MAJORVERSION \/
		n \= POOLSIZE
	) do
		error("bad image", %1);
		pr("* "); pr(p); nl();
		halt 1;
	end
	v := imagevars();
	i := 0;
	while (v[i]) do
		t.read(fd, v[i], t.bpw());
		i := i+1;
	end
	if (	t.read(fd, Car, POOLSIZE*t.bpw()) \= POOLSIZE*t.bpw() \/
		t.read(fd, Cdr, POOLSIZE*t.bpw()) \= POOLSIZE*t.bpw()
	) do
		error("bad image", %1);
		pr("* "); pr(p); nl();
		halt 1;
	end
	t.close(fd);
	if (\O_batch) do
		pr("Image loaded"); nl();
	end
end

! Print usage and fail.
usage() do
	error("Usage: sketchy [-b] [image]", %1);
	halt 1;
end

! Evaluate the command line options
getopts() do var a::TEXTLEN, i, j, k;
	O_batch := 0;
	i := 1;
	while (1) do
		k := t.getarg(i, a, TEXTLEN);
		if (k < 1 \/ a::0 \= '-') leave;
		for (j=1, k)
			ie (a::j = 'b' \/ a::j = 'B') O_batch := 1;
			else usage();
		i := i+1;
	end
	if (k > 0) str.copy(Image, a);
end

do
	init1();
	getopts();
	init2();
	if (\O_batch) do
		pr("SketchyLISP ");
		pr(RELEASE);
		pr(" (C) 2005 Nils M Holm"); nl();
	end
	if (Image::0) loadimage(Image);
	toploop();
	bye();
end

