; SketchyLISP Example Program
; Copyright (C) 2005 Nils M Holm. All rights reserved.
; See the file LICENSE of the SketchyLISP distribution
; for conditions of use.

; ---purpose---
; Evaluate a pure LISP expression in a given environment.
; --
; Primitives supported:
; BOTTOM CAR CDR COND CONS EQ? EXPLODE IMPLODE LETREC*
; LAMBDA LIST NULL? NUMBER? PAIR? PROCEDURE? QUOTE.
; --
; Internally defined symbols: #F #T.

; ---keywords---
; XEVAL function, EVAL, LISP semantics, formal semantics
; evaluate, evaluation

; ---args---
; X - s-expression to evaluate
; E - list of (name.value) pairs.

; ---example---
; ; Evaluate the Towers of Hanoi
; (define prog
;	'(letrec
;		((append2 (lambda (a b)
;			(cond ((null? a) b)
;				(#t (cons (car a) (append2 (cdr a) b))))))
;		(append3 (lambda (a b c)
;			(append2 a (append2 b c))))
;		(hanoi (lambda (a b c n) (cond ((null? n) '())
;			(#t (append3 (hanoi a c b (cdr n))
;				(cons (cons a (cons b '())) '())
;					(hanoi c b a (cdr n))))))))
;		(hanoi 'a 'b 'c '(i i i))))
;
; (xeval prog '()) => '((a b) (a c) (b c) (a b) (c a) (c b) (a b))

:l lib/list.l
:l lib/assoc.l
:l lib/pair.l
:l lib/caar.l cadr cddr
:l lib/caaar.l caadr cadar caddr
:l lib/caaaar.l cadadr caddar
:l lib/append.l
:l lib/memberp.l member?
:l lib/map.l
:l lib/not.l

; ---code---
(define xeval (lambda (x e)
	(letrec

	; Default environment
	((initial-env
		'((#f.#f)
		(#t.#t)
		(bottom.bottom)
		(call/cc.call/cc)
		(car.car)
		(cdr.cdr)
		(cond.cond)
		(cons.cons)
		(define.define)
		(eq?.eq?)
		(eval.eval)
		(explode.explode)
		(implode.implode)
		(lambda.lambda)
		(letrec.letrec)
		(list.list)
		(null?.null?)
		(pair?.pair?)
		(quote.quote)))

	; Evaluate each member of list X in environment E:
	; (M1 ... Mn) -> (eval[M1] ... eval[Mn])
	(evlist (lambda (x e)
		(cond ((null? x) '())
			(#t (cons (_eval (car x) e) (evlist (cdr x) e))))))

	; Find the value of a symbol
	(value (lambda (x e)
		(letrec ((b (assoc x e)))
			(cond ((null? b) '())
				(#t (cdr b))))))

	; Create a LETREC* environment
	(makeenv (lambda (bindings env)
		(letrec

			; Fix recursive bindings. This is a mess.
			; Lexically scoped recursive procedures are
			; cyclic structures. The construction of such
			; structures requires mutation, which purely
			; applicative languages such as SketchyLISP do
			; not provide.
			; This implementation actually cheats: To fix
			; recursive bindings, all bindings of the symbols
			; bound in LETREC* are REMOVED from the lexical
			; contexts of (mutual) recursive functions.
			; This way, the function symbols are looked up
			; in the outer environment.
			; NOTE: this trick breaks tail call optimization.
			; E is the environment to be fixed.
			((fix-rec (lambda (e)
				(letrec

					; Names of symbols bound by LETREC*
					((names (map car e))

					; Remove bindings with the given NAMES from
					; the lexical context of the function F.
					; F-ENV is the context of F.
					(remove-names (lambda (f f-env)
						(letrec

							; The function without lexical context
							((func
								(list (car f) (cadr f) (caddr f)))

							(remove (lambda (env)
								(cond ((null? env) '())
									((member? (caar env) names)
										(remove (cdr env)))
									(#t (cons (car env)
										(remove (cdr env))))))))

							(append func (list (remove f-env))))))

					; Clean up lexical contexts of lambda functions
					; bound by LETREC*. B is one binding of the context
					; created by LETREC*.
					(cleanup (lambda (b)
						(letrec ((sym (car b))
								(val (cdr b)))
							(cond ((not (pair? val)) val)
								((not (eq? (car val) 'lambda)) val)
								((null? (cdddr val)) val)
								(#t (cons sym
									(remove-names val (cadddr val))))))))

					; Fix recursive references in each binding of
					; LETREC*. B is the context created by LETREC*.
					(fix (lambda (b)
						(cond ((null? b) '())
							(#t (cons (cleanup (car b))
									(fix (cdr b))))))))

					(fix e))))

			; Evaluate and bind arguments.
			(make-bindings (lambda (bindings new-env)
				(cond ((null? bindings) new-env)
					(#t (make-bindings (cdr bindings)
						(cons (cons
								(caar bindings)
								; Evaluate arguments in the union of
								; the outer environment and the local
								; environment built so far.
								(_eval (cadar bindings)
									(append new-env env)))
							new-env)))))))

			(append (fix-rec (make-bindings bindings '())) env))))

	; Extract lexical environment of LAMBDA
	(lexical-env (lambda (x)
		(cond ((null? (cdddr x)) '())
			(#t (cadddr x)))))

	; Is symbol X bound in list A or environment E?
	(bound? (lambda (x a e)
		(cond ((eq? (assoc x e) '())
				(member? x a))
			(#t #t))))

	; Create a new symbol.value pair
	(binding (lambda (sym env)
		(list (cons sym (value sym env)))))

	; Create closure from LAMBDA function
	(make-closure (lambda (args body env)
		(letrec ((make-env (lambda (expr new-env)
				(cond ((pair? expr)
						(make-env (cdr expr)
							(make-env (car expr) new-env)))
					((null? expr)
						new-env)
					((bound? expr args new-env)
						new-env)
					(#t 
						(append (binding expr env) new-env))))))
			(list 'lambda args body (make-env body '())))))

	; Select first clause of COND that has a true predicate.
	(do-cond (lambda (clauses env)
		(cond ((null? clauses)
				(bottom 'xeval:cond-no-default))
			(#t (letrec ((tval (_eval (caar clauses) env)))
				(cond ((eq? tval #t)
						(cadar clauses))
					((eq? tval #f)
						(do-cond (cdr clauses) env))
					(#t (bottom 'xeval:cond-no-tval tval))))))))


	; Is X a procedure?
	(procedure-p (lambda (x)
		(cond ((pair? x) (eq? (car x) 'lambda))
			(#t (procedure? x)))))

	; Evaluate X in E
	(_eval (lambda (x e)
		(cond ((symbol? x)
				(value x e))
			((number? x)
				x)
			((procedure? x)
				x)
			((eq? x #t)
				#t)
			((eq? x #f)
				#f)
			((null? x)
				'())
			(#t (letrec ((fun (_eval (car x) e)))
				(cond ((eq? fun 'bottom)
						(bottom (cdr x)))
					((eq? fun 'quote)
						(cadr x))
					((eq? fun 'car)
						(car (_eval (cadr x) e)))
					((eq? fun 'cdr)
						(cdr (_eval (cadr x) e)))
					((eq? fun 'cons)
						(cons (_eval (cadr x) e) (_eval (caddr x) e)))
					((eq? fun 'pair?)
						(pair? (_eval (cadr x) e)))
					((eq? fun 'null?)
						(null? (_eval (cadr x) e)))
					((eq? fun 'eq?)
						(eq? (_eval (cadr x) e) (_eval (caddr x) e)))
					((eq? fun 'number?)
						(number? (_eval (cadr x) e)))
					((eq? fun 'procedure?)
						(procedure-p (_eval (cadr x) e)))
					((eq? fun 'cond)
						(_eval (do-cond (cdr x) e) e))
					((eq? fun 'list)
						(evlist (cdr x) e))
					((eq? fun 'implode)
						(implode (_eval (cadr x) e)))
					((eq? fun 'explode)
						(explode (_eval (cadr x) e)))
					((eq? fun 'lambda)
						(make-closure (cadr x) (caddr x) e))
					((eq? fun 'letrec)
						(_eval (caddr x) (makeenv (cadr x) e)))
					((eq? fun 'call/cc)
						(call/cc (_eval (cadr x) e)))	; cheating!
					((symbol? fun)
						(bottom (list 'xeval:non-function fun)))
					((null? fun)
						(bottom (list 'xeval:non-function (car x))))
					((eq? (car fun) 'lambda)
						(_eval (caddr fun)
							(append
								(pair (cadr fun) (evlist (cdr x) e))
								(append (lexical-env fun) e))))
					(#t (bottom (list 'xeval:non-function fun))))))))))

	(_eval x (append e initial-env)))))

